/*
 * Valve libav wrapper for h264 decoder
 */

#include <string.h>
#include "internal.h"
#include "libavutil/internal.h"
#include "avcodec.h"
#include "valve_h264.h"
#include "valve_utils.h"
#include "libavutil/intreadwrite.h"

#ifdef _WIN32
#include <windows.h>
#include <assert.h>
#include <VersionHelpers.h>

// CreateDecoder and DestroyDecoder must be called from the application main thread.
typedef H264DEC (*pfnWinH264CreateDecoder)();
typedef H264DEC (*pfnWinH264CreateDecoderD3D9)( void* pD3DDevMgrToShare );
typedef void (*pfnWinH264DestroyDecoder)(H264DEC);
typedef int (*pfnWinH264DecoderProcessInput)(H264DEC, const void *, unsigned int);
typedef int (*pfnWinH264DecoderProcessTimestamps)(H264DEC, __int64 pts, __int64 dts);
typedef void (*pfnWinH264DecoderEndOfInput)(H264DEC);
typedef int (*pfnWinH264DecoderWaitForOutput)(H264DEC, unsigned int);
typedef int (*pfnWinH264DecoderAcquireOutput)(H264DEC, const void **, unsigned int *, unsigned int *, unsigned int *, __int64 *, __int64 *);
typedef void (*pfnWinH264DecoderReleaseOutput)(H264DEC, const void *);

static pfnWinH264CreateDecoder s_pfnCreateDecoder;
static pfnWinH264CreateDecoderD3D9 s_pfnCreateDecoderD3D9;
static pfnWinH264DestroyDecoder s_pfnDestroyDecoder;
static pfnWinH264DecoderProcessInput s_pfnProcessInput;
static pfnWinH264DecoderProcessTimestamps s_pfnProcessTimestamps;
static pfnWinH264DecoderEndOfInput s_pfnEndOfInput;
static pfnWinH264DecoderWaitForOutput s_pfnWaitForOutput;
static pfnWinH264DecoderAcquireOutput s_pfnAcquireOutput;
static pfnWinH264DecoderReleaseOutput s_pfnReleaseOutput;

static int valve_h264_decode_extra_data(ValveH264Context *h);
static int valve_get_avc_nalsize( const uint8_t *buf, int buf_size, int nal_length_size );

static int load_valve_decoder(void)
{
	static int s_bAlreadyTried = 0;
	static HMODULE s_libWinH264 = NULL;

	wchar_t path[2048];
	int len;
	HMODULE hmodSelf = NULL;
	DWORD dwOldErrorMode;

	if ( s_libWinH264 )
        return 0;
	if ( s_bAlreadyTried )
		return -1;
	s_bAlreadyTried = 1;

	/* GetVersionEx may claim XP if Windows has applied compatibility hacks. look for a Vista+ function. */
	if ( !GetProcAddress( GetModuleHandleA("kernel32.dll"), "InitializeConditionVariable" ) )
		return -1;

	/* load winh264.dll wrapper from the same absolute path as libavcodec (or the exe/dll that linked it) */
	if ( FAILED( GetModuleHandleExW( GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT, (LPCWSTR)&s_libWinH264, &hmodSelf ) ) )
		return -1;

	len = GetModuleFileNameW( hmodSelf, path, sizeof( path ) );
	if ( len <= 0 || len >= 2048 )
		return -1;

	while ( len > 0 && path[len - 1] != '\\' )
		--len;
	path[len] = '\0';
	wcscat_s( path, sizeof(path)/sizeof(path[0]), L"winh264.dll" );

	dwOldErrorMode = SetErrorMode( SEM_NOOPENFILEERRORBOX | SEM_FAILCRITICALERRORS );
	s_libWinH264 = LoadLibraryW( path );

	if ( !s_libWinH264 )
	{
		path[len] = '\0';
		wcscat_s( path, sizeof( path ) / sizeof( path[0] ), L"..\\winh264.dll" );
		s_libWinH264 = LoadLibraryW( path );
	}

	SetErrorMode( dwOldErrorMode );

	if ( !s_libWinH264 )
        return -1;
    
	s_pfnCreateDecoder = (pfnWinH264CreateDecoder)GetProcAddress( s_libWinH264, "WinH264CreateDecoder" );
	s_pfnCreateDecoderD3D9 = (pfnWinH264CreateDecoderD3D9)GetProcAddress( s_libWinH264, "WinH264CreateDecoderD3D9" );
    s_pfnDestroyDecoder = (pfnWinH264DestroyDecoder)GetProcAddress( s_libWinH264, "WinH264DestroyDecoder" );
    s_pfnProcessInput = (pfnWinH264DecoderProcessInput)GetProcAddress( s_libWinH264, "WinH264DecoderProcessInput" );
    s_pfnProcessTimestamps = (pfnWinH264DecoderProcessTimestamps)GetProcAddress( s_libWinH264, "WinH264DecoderProcessTimestamps" );
    s_pfnEndOfInput = (pfnWinH264DecoderEndOfInput)GetProcAddress( s_libWinH264, "WinH264DecoderEndOfInput" );
    s_pfnWaitForOutput = (pfnWinH264DecoderWaitForOutput)GetProcAddress( s_libWinH264, "WinH264DecoderWaitForOutput" );
    s_pfnAcquireOutput = (pfnWinH264DecoderAcquireOutput)GetProcAddress( s_libWinH264, "WinH264DecoderAcquireOutput" );
    s_pfnReleaseOutput = (pfnWinH264DecoderReleaseOutput)GetProcAddress( s_libWinH264, "WinH264DecoderReleaseOutput" );
    if ( !s_pfnCreateDecoder || !s_pfnDestroyDecoder || !s_pfnProcessInput || !s_pfnEndOfInput || !s_pfnWaitForOutput || !s_pfnAcquireOutput || !s_pfnReleaseOutput || !s_pfnProcessTimestamps )
    {
        FreeLibrary( s_libWinH264 );
        s_libWinH264 = NULL;
        return -1;
    }
    
    return 0;
}

int valve_load_and_check_winh264_dll(void)
{
	H264DEC dec;
	if ( load_valve_decoder() < 0 )
		return -1;
	dec = s_pfnCreateDecoder();
	if ( !dec )
		return -1;
	s_pfnDestroyDecoder( dec );
	return 0;
}
#endif

static void valve_copy_plane( uint8_t *__restrict dst, int dst_stride, uint8_t *__restrict src, int src_stride, int width, int height )
{
	for ( int i = 0; i < height; i++ )
	{
		memcpy( dst, src, width );
		dst += dst_stride;
		src += src_stride;
	}
}

static void valve_extract_nv12_planes( uint8_t *__restrict dstu, int dstu_stride, uint8_t *__restrict dstv, int dstv_stride, uint8_t *__restrict src, int src_stride, int width, int height )
{
	for ( int i = 0; i < height; i++ )
	{
		int x = width;
		while ( x > 7 )
		{
			x -= 8;
			dstv[x+7] = src[2*x+15];
			dstv[x+6] = src[2*x+13];
			dstv[x+5] = src[2*x+11];
			dstv[x+4] = src[2*x+9];
			dstv[x+3] = src[2*x+7];
			dstv[x+2] = src[2*x+5];
			dstv[x+1] = src[2*x+3];
			dstv[x+0] = src[2*x+1];
			dstu[x+7] = src[2*x+14];
			dstu[x+6] = src[2*x+12];
			dstu[x+5] = src[2*x+10];
			dstu[x+4] = src[2*x+8];
			dstu[x+3] = src[2*x+6];
			dstu[x+2] = src[2*x+4];
			dstu[x+1] = src[2*x+2];
			dstu[x+0] = src[2*x+0];
		}
		while ( x > 0 )
		{
			--x;
			dstv[x] = src[2*x+1];
			dstu[x] = src[2*x+0];
		}
		dstu += dstu_stride;
		dstv += dstv_stride;
		src += src_stride;
	}
}

static int valve_get_avc_nalsize( const uint8_t *buf, int buf_size, int nal_length_size )
{
    int i, nalsize = 0;

    if ( buf_size < nal_length_size )
        return -1;
        
    for ( i = 0; i < nal_length_size; i++ )
        nalsize = (nalsize << 8) | buf[ i ];
        
    if (nalsize <= 0 )
        return -1;
        
    return nalsize;
}

static int advance_buffer( const uint8_t **ppData, int *pLength, int amount )
{
    if ( *pLength < amount )
        return -1;
        
    *ppData = *ppData + amount;
    *pLength = *pLength - amount;
    
    return 0;
}

static int advance_buffer_not_empty( const uint8_t **ppData, int *pLength, int amount )
{
    if ( advance_buffer( ppData, pLength, amount ) < 0 )
        return -1;
        
    return (*pLength > 0);
}

#define RET_FAIL( c ) { if ( c < 0 ) return -1; }

#ifdef _WIN32
static int valve_h264_decode_extra_data(ValveH264Context *h)
{
    AVCodecContext *avctx = h->avctx;
    unsigned char *pData = avctx->extradata;
    int data_length = avctx->extradata_size;
    int i = 0;
    int count = 0;
    unsigned char *pNal = NULL;
    int size = 0;
    
	if ( !h->hDecoder )
		return -1;

    if ( data_length == 0 || !pData )
	{
		h->is_avc = 0;
        return 0;
	}
        
    if ( pData[0] != 1)
        return -1;

	h->is_avc = 1;

    // skip sps start
    RET_FAIL( advance_buffer_not_empty( &pData, &data_length, 4 ) );
    h->nal_length_size = (*pData & 0x03) + 1;
    RET_FAIL( advance_buffer_not_empty( &pData, &data_length, 1 ) );
    count = (*pData & 0x1f );
    RET_FAIL( advance_buffer_not_empty( &pData, &data_length, 1 ) );

    pNal = NULL;
    size = 0;
    for ( i = 0; i < count; i++ )
    {
        size = AV_RB16( pData );
        RET_FAIL( advance_buffer_not_empty( &pData, &data_length, 2 ) );
                  
        pNal = pData;
        RET_FAIL( advance_buffer_not_empty( &pData, &data_length, size ) );
    }
    
    if ( !pNal || size <= 0 )
        return -1;
    
    s_pfnProcessInput( h->hDecoder, pNal, size );
    
    // pps
    count = *pData;
    RET_FAIL( advance_buffer_not_empty( &pData, &data_length, 1 ) );
    
    pNal = NULL;
    size = 0;
    for ( i = 0; i < count; i++ )
    {
        if ( data_length <= 0 )
            return -1;
        
        size = AV_RB16( pData );
        RET_FAIL( advance_buffer_not_empty( &pData, &data_length, 2 ) );
        
        pNal = pData;
        RET_FAIL( advance_buffer( &pData, &data_length, size ) );
    }
    
    if ( !pNal || size <= 0 )
        return -1;
    
    s_pfnProcessInput( h->hDecoder, pNal, size );
    
    return 0;
}

#endif


// exported
static int valve_h264_decode_init( AVCodecContext *avctx )
{
#ifdef _WIN32
	ValveH264Context *h = avctx->priv_data;
	if ( load_valve_decoder() < 0 )
	{
		return -1;
	}

	h->avctx = avctx;
	h->hDecoder = 0;
	if ( s_pfnCreateDecoderD3D9 )
	{
		h->output_is_nv12 = 1;
		h->hDecoder = s_pfnCreateDecoderD3D9( (void*)0 );
	}
	if ( !h->hDecoder )
	{
		h->output_is_nv12 = 0;
		h->hDecoder = s_pfnCreateDecoder();
	}
	h->nal_length_size = 0;
	h->is_avc = -1; // unknown. will sniff on first frame decode.
	avctx->pix_fmt = AV_PIX_FMT_YUV420P;

	return h->hDecoder ? 0 : -1;
#else
	return -1;
#endif
}

// exported
static int valve_h264_decode_frame( AVCodecContext *avctx, void *data, int *got_frame, AVPacket *avpkt )
{
#ifdef _WIN32
    const uint8_t *buf = avpkt->data;
    int buf_size = avpkt->size;
    ValveH264Context *h = avctx->priv_data;
    AVFrame *frame = data;
    int wait_ms = 0;
    *got_frame = 0;
    
	if ( !h->hDecoder )
	{
		return -1;
	}

	if ( h->is_avc == -1 )
	{
		// we will have extradata if packet is in AVC format; if so, this will set h->is_avc
		if ( valve_h264_decode_extra_data( h ) < 0 )
		{
			h->is_avc = -2; // stop trying to sniff extradata on parse error.
		}
	}
	if ( h->is_avc < 0 )
	{
		return -1;
	}

	// if we're being drained, make sure we signal end of input. it's ok to signal multiple times.
	if ( avpkt->data == NULL && avpkt->size == 0 )
	{
		s_pfnEndOfInput( h->hDecoder );
	}
	else
	{
		s_pfnProcessTimestamps( h->hDecoder, avpkt->pts, avctx->reordered_opaque );
	}

	if ( h->is_avc > 0 )
	{
		// AVC - parse out each length-prefixed NAL till end of buffer    
		if ( h->nal_length_size == 0 )
		{
			return -1;
		}
		while ( buf && buf_size > 0 )
		{
			int nal_size = 0;
			nal_size = valve_get_avc_nalsize( buf, buf_size, h->nal_length_size );
			if ( nal_size <= 0 )
			{
				//OutputDebugString( "**** NAL ERROR\n" );
				return -1;
			}
            
			RET_FAIL( advance_buffer_not_empty( &buf, &buf_size, h->nal_length_size ) );
        
			s_pfnProcessInput( h->hDecoder, buf, nal_size );
        
			RET_FAIL( advance_buffer( &buf, &buf_size, nal_size ) );
		}
	}
	else
	{
		// annex b - scan for NAL start codes and slice into input packets accordingly
		const uint8_t *nalstart = 0;
		unsigned int x = 0;
		while ( buf && buf_size > 0 )
		{
			--buf_size;
			x = (x * 256 + (unsigned char)*buf++) & 0xffffff;
			if ( x == 1 ) // found a 3-byte start code: bytes 00 00 01
			{
				if ( nalstart )
					s_pfnProcessInput( h->hDecoder, nalstart, buf - nalstart - 3 );
				nalstart = buf;
			}
		}
		if ( nalstart && buf > nalstart ) // final NAL goes to end of packet
			s_pfnProcessInput( h->hDecoder, nalstart, buf - nalstart );
	}
    
    //OutputDebugString( "**** ADDED FRAME\n" );
    
    // wait for input. If we have already returned a frame, *really* want to wait for next frame
    wait_ms = 5;
    if ( h->decoded_frames > 0 )
        wait_ms = 1000;
        
    if ( s_pfnWaitForOutput( h->hDecoder, wait_ms ) == WINH264_OK )
    {
        void *buffer;
        unsigned int width = 0;
        unsigned int height = 0;
        unsigned int stride = 0;
        __int64 timestamp = 0;
        __int64 opaque = 0;
        
        int res = s_pfnAcquireOutput( h->hDecoder, &buffer, &stride, &width, &height, &timestamp, &opaque );
        if ( res == WINH264_ERR_WOULD_BLOCK )
        {
            //OutputDebugString( "**** NO FRAME\n" );
            return avpkt->size;
        }
        
        if ( res != WINH264_OK )
        {
            //OutputDebugString( "**** ERROR\n" );
            return -1;
        }
        
        //OutputDebugString( "**** GOT FRAME\n" );
        if ( width < avctx->width || height < avctx->height || ff_get_buffer( avctx, frame, 0 ) < 0 )
        {
            //OutputDebugString( "**** INVALID DIMENSIONS\n" );
            s_pfnReleaseOutput( h->hDecoder, buffer );
            return -1;
        }
        
        h->decoded_frames++;

		if ( h->output_is_nv12 )
		{
			/* convert to from Y/UV-packed planes to Y/U/V planes */
			uint8_t *pYPlane = buffer;
			uint8_t *pUVPlane = pYPlane + (stride * height);
			valve_copy_plane( frame->data[0], frame->linesize[0], pYPlane, stride, avctx->width, avctx->height );
			valve_extract_nv12_planes( frame->data[1], frame->linesize[1], frame->data[2], frame->linesize[2], pUVPlane, stride, avctx->width / 2, avctx->height / 2 );
		}
		else
		{
			uint8_t *pYPlane = buffer;
			uint8_t *pUPlane = pYPlane + (stride * height);
			uint8_t *pVPlane = pUPlane + (stride * height / 4);
			valve_copy_plane( frame->data[0], frame->linesize[0], pYPlane, stride, avctx->width, avctx->height );
			valve_copy_plane( frame->data[1], frame->linesize[1], pUPlane, stride / 2, avctx->width / 2, avctx->height / 2 );
			valve_copy_plane( frame->data[2], frame->linesize[2], pVPlane, stride / 2, avctx->width / 2, avctx->height / 2 );
		}
        
        frame->width = avctx->width;
        frame->height = avctx->height;
        frame->format = avctx->pix_fmt;
        frame->pts = AV_NOPTS_VALUE;
        frame->pkt_pts = timestamp;
        frame->pkt_dts = AV_NOPTS_VALUE;
		frame->reordered_opaque = opaque;
                
        s_pfnReleaseOutput( h->hDecoder, buffer );
        
        *got_frame = 1;
    }
    else
    {
        //OutputDebugString( "**** WAIT FIAL\n" );
    }
    
    return avpkt->size;
#else
    return -1;
#endif
}

// exported
static int valve_h264_decode_end( AVCodecContext *avctx )
{
#ifdef _WIN32
    ValveH264Context *h = avctx->priv_data;

    if ( h->hDecoder )
    {
        s_pfnDestroyDecoder( h->hDecoder );
        h->hDecoder = NULL;
    }
#endif

    return 0;
}

// exported. flush after seek by destroying, recreating, reprocessing SPS/PPS
static void valve_h264_flush( AVCodecContext *avctx )
{
#ifdef _WIN32
	ValveH264Context *h = avctx->priv_data;
	H264DEC hDecOld = h->hDecoder;
	h->hDecoder = 0;
	if ( s_pfnCreateDecoderD3D9 )
	{
		h->output_is_nv12 = 1;
		h->hDecoder = s_pfnCreateDecoderD3D9( (void*)0 );
	}
	if ( !h->hDecoder )
	{
		h->output_is_nv12 = 0;
		h->hDecoder = s_pfnCreateDecoder();
	}
	s_pfnDestroyDecoder( hDecOld );
#endif
}

static int valve_h264_enabled(void)
{
	return use_valve_h264();
}

static const AVProfile profiles[] = {
    { FF_PROFILE_H264_BASELINE,             "Baseline"              },
    { FF_PROFILE_H264_CONSTRAINED_BASELINE, "Constrained Baseline"  },
    { FF_PROFILE_H264_MAIN,                 "Main"                  },
    { FF_PROFILE_H264_EXTENDED,             "Extended"              },
    { FF_PROFILE_H264_HIGH,                 "High"                  },
    { FF_PROFILE_H264_HIGH_10,              "High 10"               },
    { FF_PROFILE_H264_HIGH_10_INTRA,        "High 10 Intra"         },
    { FF_PROFILE_H264_HIGH_422,             "High 4:2:2"            },
    { FF_PROFILE_H264_HIGH_422_INTRA,       "High 4:2:2 Intra"      },
    { FF_PROFILE_H264_HIGH_444,             "High 4:4:4"            },
    { FF_PROFILE_H264_HIGH_444_PREDICTIVE,  "High 4:4:4 Predictive" },
    { FF_PROFILE_H264_HIGH_444_INTRA,       "High 4:4:4 Intra"      },
    { FF_PROFILE_H264_CAVLC_444,            "CAVLC 4:4:4"           },
    { FF_PROFILE_UNKNOWN },
};

AVCodec ff_valve_h264_decoder = {
    .name                  = "valve_h264",
    .long_name             = NULL_IF_CONFIG_SMALL("H.264 / AVC / MPEG-4 AVC / MPEG-4 part 10"),
    .type                  = AVMEDIA_TYPE_VIDEO,
    .id                    = AV_CODEC_ID_H264,
    .priv_data_size        = sizeof(ValveH264Context),
    .enabled               = valve_h264_enabled,
    .init                  = valve_h264_decode_init,
    .close                 = valve_h264_decode_end,
    .decode                = valve_h264_decode_frame,
    .capabilities          = /*CODEC_CAP_DRAW_HORIZ_BAND |*/ CODEC_CAP_DR1 | CODEC_CAP_DELAY /*| CODEC_CAP_SLICE_THREADS | CODEC_CAP_FRAME_THREADS*/,
    .flush                 = valve_h264_flush,
    /*.init_thread_copy      = ONLY_IF_THREADS_ENABLED(valve_h264_init_thread_copy),
    .update_thread_context = ONLY_IF_THREADS_ENABLED(valve_h264_update_thread_context),*/
    .profiles              = NULL_IF_CONFIG_SMALL(profiles),
};
